from Products.CMFCore.utils import getToolByName
from Products.Archetypes.public import listTypes
from Products.Archetypes.Extensions.utils import installTypes, install_subskin
from Products.PloneWurfl.PloneWurflTool import \
    PloneWurflTool
from Products.PloneWurfl.config import PROJECTNAME, GLOBALS
from Products.PloneWurfl.MobileDevice import schema as MobileDevice_schema
from StringIO import StringIO
from Products.ATSchemaEditorNG.config import TOOL_NAME as SCHEMA_EDITOR

def install(self):
    out = StringIO()

    installTypes(self, out, listTypes(PROJECTNAME), PROJECTNAME)

    install_subskin(self, out, GLOBALS)

    installed_dependencies = install_dependencies(self, out)

    install_tools(self, out)

    # Store fact that we installed PloneAzax
    if 'PloneAzax' in installed_dependencies:
        tool = getToolByName(self, 'portal_wurfl')
        tool.we_installed_ploneazax = True
        
    # KSS sheet
    pcss = getToolByName(self, 'portal_css')
    pcss.manage_addStylesheet(id='plonewurfl.kss',
        rel='k-stylesheet',
        rendering='link',
        enabled=True)
    pcss.manage_addStylesheet(id='plonewurfl.css',
        rel='stylesheet',
        rendering='import',
        enabled=True)
        
    # Create External method
    if not hasattr(self, 'wurfl_import_from_xml'):
        em = self.manage_addProduct['ExternalMethod']
        em.manage_addExternalMethod('wurfl_import_from_xml', 'wurfl_import_from_xml', 
            'PloneWurfl.wurfl_import_from_xml', 'run')

    out.write("Successfully installed %s." % PROJECTNAME)
    return out.getvalue()

def install_dependencies(portal, out, required=1):
    qi=getToolByName(portal, 'portal_quickinstaller', None)
    if qi is None and required:
        raise RuntimeError, (
            "portal_quickinstaller tool could not be found, and it is "
            "required to install PloneWurfl's dependencies")

    ret = []
    if not qi.isProductInstalled('ATExtensions'):
        qi.installProduct('ATExtensions')
        ret.append('ATExtensions')
        print >>out, 'Installing ATExtensions'

    if not qi.isProductInstalled('ATSchemaEditorNG'):
        qi.installProduct('ATSchemaEditorNG')
        ret.append('ATSchemaEditorNG')
        print >>out, 'Installing ATSchemaEditorNG'
        
    if not qi.isProductInstalled('PloneAzax'):
        qi.installProduct('PloneAzax')
        ret.append('PloneAzax')
        print >>out, 'Installing PloneAzax'

    return ret

def install_tools(portal, out):
    # Register schema's with ATSchemaEditor
    SE = getToolByName(portal, SCHEMA_EDITOR)
    if not SE.atse_isSchemaRegistered('MobileDevice'):
        SE.atse_registerSchema('MobileDevice', MobileDevice_schema, 
            filtered_schemas=('metadata', ))

    # install wurfl tool
    if not hasattr(portal, 'portal_wurfl'):
        portal.manage_addProduct[PROJECTNAME].manage_addTool(
            PloneWurflTool.meta_type, None)
        print >> out, "Successfully installed Plone Wurfl Tool."
