def run(self, xmlfile=None, update_existing_devices=False, schema_only=False, debug=True, REQUEST=None):
    from Products.CMFCore.utils import getToolByName
    from xml.parsers.expat import ParserCreate
    from ZPublisher.HTTPRequest import record
    from logging import getLogger
    import urllib2
    import zipfile
    from cStringIO import StringIO

    def plone_log(summary='', text=''):
        log = getLogger('Plone')
        log.info('Debug: %s \n%s', summary, text)

    request = getattr(self, 'REQUEST', None)

    se = getToolByName(self, 'schema_editor_tool')
    pc = getToolByName(self, 'portal_catalog')
       
    # Are we called on a root MobileDevice?
    if not ((getattr(self, 'portal_type', '') == 'MobileDevice') \
        and  (getattr(self.aq_parent, 'portal_type', '') != 'MobileDevice')):
        raise "[wurfl_import_from_xml] - must be called on a root MobileDevice!"

    # Read from file object
    if debug:
        fp = open('/tmp/wurfl.xml')
        try:
            xml = fp.read()
        finally:
            fp.close()
    elif xmlfile is not None:
        xmlfile.seek(0)
        xml = xmlfile.read()
        if not xml:
            # Get file from web
            f = urllib2.urlopen('http://wurfl.sourceforge.net/wurfl.zip')
            data = f.read()
            zf = zipfile.ZipFile(StringIO(data), 'r')
            xml = zf.read('wurfl.xml')
            
    if not xml:
        raise "[wurfl_import_from_xml] - no XML to parse"

    def fieldprops_to_dict(field):
        # Return a dictionary for use in updating the field
        field_data = {
                       'type': 'StringField',
                        'schemata': self.state_group, 
                        'name': self.state_capability,
                        'widget': 'String',
                        'label': field.widget.label, 
                        'storage': field.storage.getName().replace('Storage', ''),
                        'visible_edit': 1,
                        'visible_view': 1,
                        'mutator': field.mutator,
                      }
        return field_data
    
    # Parse XML and update schema
    self.state_device_id = ''
    self.state_group = ''
    self.state_capability = ''
    self.schema_modified = False
    
    # The XML is inorder, so we'll need to sort it. Also, 
    # we can't keep the objects in a list because of memory
    # considerations. So... sort the device_ids beforehand.
    self.device_xml_blocks = []              
    self.xml = {}
    def firstpass_start_element(name, attrs):

        # Device
        if name == 'device':
            self.state_device_id = str(attrs['id'])
            self.xml = {
                'device_id': self.state_device_id,
                'fall_back': str(attrs['fall_back']),
                'user_agent': str(attrs['user_agent']),
                'actual_device_root': str(attrs.get('actual_device_root', '')) in ('true','True'),
                'groups': []
                }

        # Group
        elif name == 'group':
            self.state_group = str(attrs['id'])
            self.xml['groups'].append(
                {
                    'group_id':self.state_group, 
                    'capabilities':[]
                }
                )
             
            # Use generic device to bootstrap variable schema
            if self.state_device_id == 'generic':
                # Create schemata if it does not exist
                if self.state_group not in se.atse_getSchemataNames('MobileDevice'):
                    se.atse_addSchemata('MobileDevice', '', self.state_group)  

        # Capability
        elif name == 'capability':
            self.state_capability = str(attrs['name'])
            value = str(attrs['value'])
            self.xml['groups'][-1]['capabilities'].append(
                {
                    'name':self.state_capability,
                    'value':value,
                }
                )
            
            field_data = {}
                
            # Determine field type
            field_type = 'StringField'
            field_type_short = 'string'
            field_widget = 'String'
            field_default = ''
            field_vocabulary = []
            if value.lower() in ('true','false','True','False'):                                                
                field_type = 'IntegerField'
                field_type_short = 'integer'
                field_widget = 'Radio'
                field_default = "-1"
                field_vocabulary = ['1|Yes', '0|No','-1|Inherit']
            else:
                try: 
                    i = int(value)                            
                    field_type = 'IntegerField'
                    field_type_short = 'integer'
                except:                        
                    pass

            # Try and find the field
            try:
                fld = se.atse_getField('MobileDevice', self.state_capability)

                # The field is registered with schema editor. 
                # We may encounter a case where the field is currently of type
                # integer or boolean, but the current capability value is
                # a string. In  such a case change the type. 
                # This problem arises because there is no sufficient DTD for the 
                # Wurfl XML file.
                if (fld.type != field_type_short) and (field_type_short == 'string'):
                    plone_log("[wurfl_import_from_xml - modify field %s (value=%s)" % (self.state_capability, value))
                    field_data = fieldprops_to_dict(fld)                   
                    field_data['type'] = field_type
                    field_data['widget'] = field_widget                        
                        
            except KeyError:        
                # Field does not yet exist. Create.
                plone_log("[wurfl_import_from_xml - create field %s" % self.state_capability)
                se.atse_addField('MobileDevice', self.state_group, '', self.state_capability)
                fld = se.atse_getField('MobileDevice', self.state_capability)
                 
                field_data = fieldprops_to_dict(fld)                   
                field_data['type'] = field_type
                field_data['widget'] = field_widget                        
                field_data['default'] = field_default
                field_data['vocabulary'] = field_vocabulary

            # Anything to do?
            if field_data.keys():
            
                # Set widget size
                if field_data['type'] == 'StringField':
                    field_data['widgetsize'] = 60
                elif field_data['type'] == 'IntegerField':
                    field_data['widgetsize'] = 10

                # Make dictionary keys attributes of a record
                fd = record()
                for k, v in field_data.items():
                    setattr(fd, k, v)

                #se.atse_update('MobileDevice', '', fd, request)
                fake_request = record()
                setattr(fake_request, 'form', {})
                se.atse_update('MobileDevice', '', fd, fake_request)
                self.schema_modified = True
                                         
    def firstpass_end_element(name):              
        if name == 'device':             
            # Add to buffer and clear var
            self.device_xml_blocks.append(self.xml)
            self.xml = {}
    
    # First pass
    p = ParserCreate()
    p.StartElementHandler = firstpass_start_element
    p.EndElementHandler = firstpass_end_element
    p.Parse(xml, 1)
               
    if self.schema_modified:
            # Get rid of the implicitly created _default field for every schemata
            for schemata_name in se.atse_getSchemataNames('MobileDevice'):
                # Only delete default if there are other fields in schemata.
                # We don't want to break ATSchemaEditorNG
                if len(se.atse_getSchemata('MobileDevice', schemata_name).fields()):
                    try:
                        fld = se.atse_getField('MobileDevice', schemata_name+'_default')
                        se.atse_delField('MobileDevice', '', schemata_name+'_default')
                    except KeyError:
                        pass
            
            # Update
            se.atse_updateManagedSchema('MobileDevice', '', update_all=True)
            
    if schema_only:
        return 'done'

    # Anything to do?
    if not self.device_xml_blocks:
        return 'done'

    # Sort self.device_xml_blocks.    
    root_xml = self.device_xml_blocks[0]    
    self.sorted_device_xml_blocks = [root_xml]
    self.sorted_device_ids = [root_xml['device_id']]
    plone_log("[wurfl_import_from_xml] - recurse...")
    def recurser(node):
        for xml in self.device_xml_blocks:
            if xml['fall_back'] == node['device_id']:
                self.sorted_device_xml_blocks.append(xml)
                self.sorted_device_ids.append(node['device_id'])
                recurser(xml)
        
    recurser(root_xml)
  
    # In case of an XML diff we will have "orphaned" devices.
    # Append these to sorted_device_xml_blocks
    for xml in self.device_xml_blocks:
        if not xml['device_id'] in self.sorted_device_ids:
            self.sorted_device_xml_blocks.append(xml)
    
    # Verify the sort
    plone_log("[wurfl_import_from_xml] - verify sort...")
    i = 0
    for xml in self.sorted_device_xml_blocks:
        j = 0
        for x in self.sorted_device_xml_blocks:
            if x['device_id'] == xml['fall_back']:
                if j > i:
                    raise "Bad sort. Device %s precedes its parent %s." % (xml['device_id'], xml['fall_back'])
            j += 1    
        i += 1

    # Now create the object tree
    counter = 1
    total = len(self.sorted_device_xml_blocks)
    number_created = 0
    number_updated = 0
    if total:
        plone_log("[wurfl_import_from_xml] - creating/updating %s devices..." % total)
    for xml in self.sorted_device_xml_blocks:
        if counter % 10 == 0:
            plone_log("[wurfl_import_from_xml] - progress %.2f%%" % (counter*100.0/total))
        if counter % 50 == 0:
            get_transaction().commit()
        counter += 1

        is_new_object = False
        if xml['device_id'] != 'generic':                
            # If an object with OriginalDeviceId already exists 
            # then nothing is created. 
            brains = pc(portal_type='MobileDevice', path="/".join(self.getPhysicalPath()),
                getOriginalDeviceId=xml['device_id'])
            if brains:                
                # Found existing object
                if not update_existing_devices:
                    plone_log("[wurfl_import_from_xml] - device with id %s already exists!" % xml['device_id'])
                    continue
                ob = brains[0].getObject()
                number_updated += 1
                 
            else:
                # Find the parent object
            
                # 'self' should be THE root MobileDevice, but it may not be called
                # generic in case the system has many device trees.
                fall_back = xml['fall_back']
                if fall_back == 'generic':
                    fall_back = self.getId()
                    
                brains = pc(portal_type='MobileDevice', path="/".join(self.getPhysicalPath()),
                    getId=fall_back)
                if not brains:
                    raise "Unable to locate parent for %s" % xml['device_id']
                parent = brains[0].getObject()
            
                # device_id may be invalid for the ZODB. Sanitize.
                device_id = xml['device_id']
                device_id = device_id.lstrip('_')
                if device_id.endswith('__'):
                    device_id = device_id.rstrip('_') 
                parent.invokeFactory('MobileDevice', device_id)
                ob = parent[device_id]
                ob.setOriginalDeviceId(xml['device_id'])
                is_new_object = True
                number_created += 1
        else:
            ob = self
            number_updated += 1
        
        ob.setUserAgent(xml['user_agent'])
        # If this field was possibly flagged as modified then unflag 
        if not is_new_object:
            ob.unFlagModified('user_agent')
            
        ob.setIsActualDeviceRoot(xml['actual_device_root'])
        # If this field was possibly flagged as modified then unflag 
        if not is_new_object:
            ob.unFlagModified('IsActualDeviceRoot')

        for group in xml['groups']:
            for cap in group['capabilities']:                
                name = cap['name']
                value = cap['value']
               
                # If this field was possibly flagged as modified then unflag 
                ob.unFlagModified(name)

                # Booolean is stored as integer because we need three states
                # and require processing here.
                vlow = value.lower()
                if vlow in ('true','false'):
                    if vlow == 'false':
                        value = 0
                    elif vlow == 'true':
                        value = 1
                    else:
                        value = -1
                    
                ob.getField(name).set(ob, value)
                
        ob.reindexObject()
        
    if total:    
        plone_log("[wurfl_import_from_xml] - progress 100.0%")
        plone_log("[wurfl_import_from_xml] - created %s devices" % number_created)
        plone_log("[wurfl_import_from_xml] - updated %s devices" % number_updated)
 
  
    return 'done'
