from AccessControl import ClassSecurityInfo
from Products.CMFCore import CMFCorePermissions
from Products.CMFCore.utils import getToolByName
from Products.Archetypes.public import *
from Products.ATSchemaEditorNG.config import TOOL_NAME as SCHEMA_EDITOR
from Products.ATSchemaEditorNG.config import ATSE_MANAGED_NONE, \
     ATSE_MANAGED_FULL
from Products.ATSchemaEditorNG.ManagedSchema import ManagedSchema
from Products.PloneWurfl.config import PROJECTNAME
import urllib2
from cStringIO import StringIO

schema=BaseSchema.copy() + ManagedSchema((
    StringField('OriginalDeviceId',        
        default="generic",
        widget=StringWidget(), 
        index='FieldIndex:brains'
    ),
    LinesField('ModifiedFields',
        widget=LinesWidget(visible={'edit':False, 'view':True}, 
            description="Fields that have been modified. This is used to generate XML patches."),
    ),
    StringField('UserAgent',        
        widget=StringWidget(),
        index='FieldIndex:brains'
    ),
    StringField('brand_name',
        schemata='product_info',
        widget=StringWidget(),
    ),
    StringField('model_name',
        schemata='product_info',
        widget=StringWidget(),
    ),    
    ComputedField('HasModifiedFields',
        expression='len(context.getModifiedFields()) > 0',
        widget=BooleanWidget(visible=False),
        index='FieldIndex'
    ),
    BooleanField('IsActualDeviceRoot',
        widget=BooleanWidget(label="Is root device"),
        index='FieldIndex'
    ), 

    # Following three fields only applicable to root device
    StringField('SynchMasterServerUrl',               
        widget=StringWidget(visible=False), 
    ),
    StringField('SynchMasterServerUsername',               
        widget=StringWidget(visible=False), 
    ),
    StringField('SynchMasterServerPassword',               
        widget=StringWidget(visible=False), 
    ),
),
)

IdField = schema['id']
IdField.atse_managed = ATSE_MANAGED_NONE
TitleField = schema['title']
TitleField.required = 0
TitleField.visible = {'edit':False, 'view':True}
TitleField.atse_managed = ATSE_MANAGED_NONE

class MobileDevice(VariableSchemaSupport, OrderedBaseFolder):
    security = ClassSecurityInfo()
    archetype_name = 'MobileDevice'
    schema = schema
    allowed_content_types = ('MobileDevice',)
    #content_icon = 'MobileDevice.png'
    immediate_view = 'base_view'
    global_allow = 1
    filter_content_types = 0
    _at_rename_after_creation = True
    
    actions = (
        {'id': 'importxml',
         'name': 'Import / Synch',
         'action': 'string:${object_url}/wurfl_import_xml',
         'condition': "python:object.aq_parent.portal_type != 'MobileDevice'",
         'permissions': (CMFCorePermissions.ModifyPortalContent,),
        },
        {'id': 'wurfl_tree',
         'name': 'Tree',
         'action': 'string:${object_url}/wurfl_tree_edit',
         'permissions': (CMFCorePermissions.ModifyPortalContent,),
        },
        {'id': 'wurfl_xml_diff',
         'name': 'XML Diff',
         'action': 'string:${object_url}/wurfl_xml_diff',
         'permissions': (CMFCorePermissions.View,),
        },
        {'id': 'wurfl_xml_diff_tree',
         'name': 'XML Diff for Entire Tree',
         'action': 'string:${object_url}/wurfl_xml_diff_tree',
         'condition': "python:object.aq_parent.portal_type != 'MobileDevice'",
         'permissions': (CMFCorePermissions.View,),
        },
    )

    def isFieldThreeStateBoolean(self, fieldname):
        fld = self.getField(fieldname)
        return (fld.type == 'integer') and (fld.default in (-1, "-1")) \
            and fld.vocabulary
    
    def getAcquiredValue(self, key):        
        value = self.get(key)

        inherit_fields = []
        # Sorry for the following line
        dummy = [inherit_fields.extend([f.getName() for f in fieldset.fields()]) \
            for k,fieldset in self.Schemata().items() if k not in ('default','metadata')]
       
        #fld = self.getField(key)
        #self.plone_log("default for %s = %s which is a %s" % (key, fld.default, type(fld.default)))
        
        is_three_state_boolean = self.isFieldThreeStateBoolean(key)
        if ( (is_three_state_boolean and (value == -1)) or (value is None) ) \
            and (key in inherit_fields):
            parent = self.aq_parent
            v = None
            while getattr(parent, 'portal_type', '') == 'MobileDevice':
                v = parent.get(key)
                
                if is_three_state_boolean and (v != -1):
                    # Return 0 or 1 as True or False
                    return v == 1
                    
                elif v is not None:
                    return v
                    
                parent = parent.aq_parent

        if is_three_state_boolean:
            return value == 1
        else:
            return value
        
    security.declareProtected(CMFCorePermissions.View, 'getSchema')
    def getSchema(self):
        # we are not wrapped yet
        if not getattr(self, 'aq_parent', None):
            return schema
        se = getToolByName(self, SCHEMA_EDITOR)
        return se.atse_getSchemaById('MobileDevice')

    def getBrandAndModel(self):
        """ Compute Title from other fields """
        if (self.getOriginalDeviceId() == 'generic') \
            and hasattr(self, 'aq_parent') \
            and (getattr(self.aq_parent, 'portal_type', '') != 'MobileDevice'):
            return "Root Mobile Device"
        s = '%s %s' % (self.getBrand_name(), self.getModel_name())
        if not s.strip():
            return self.getId()
        return s

    Title = getBrandAndModel

    security.declareProtected(CMFCorePermissions.ManagePortal, 'doUpdateSchemaFromXML')
    def doUpdateSchemaFromXML(self, wurflfile=None, REQUEST=None):
        """Update schema"""
        res = self.wurfl_import_from_xml(self, wurflfile, schema_only=True, 
            debug=False)
        if res:
            if REQUEST is not None:
                REQUEST.RESPONSE.redirect(self.absolute_url() 
                    + '/wurfl_import_xml?portal_status_message=Success')
            return True
            
        if REQUEST is not None:
            REQUEST.RESPONSE.redirect(self.absolute_url() 
                + '/wurfl_import_xml?portal_status_message=Operation%20failed')
        return False
        
    security.declareProtected(CMFCorePermissions.ManagePortal, 'doUpdateDeviceTreeXML')
    def doUpdateDeviceTreeXML(self, wurflfile=None, update_existing_devices=False, REQUEST=None):
        """Update device tree"""
        res = self.wurfl_import_from_xml(self, wurflfile, schema_only=False, 
            update_existing_devices=update_existing_devices, debug=False)
        if res:
            if REQUEST is not None:
                REQUEST.RESPONSE.redirect(self.absolute_url() 
                    + '/wurfl_import_xml?portal_status_message=Success')
            return True
            
        if REQUEST is not None:
            REQUEST.RESPONSE.redirect(self.absolute_url() 
                + '/wurfl_import_xml?portal_status_message=Operation%20failed')
        return False
        
    security.declareProtected(CMFCorePermissions.ManagePortal, 'doUpdateDeviceTreeFromMasterServer')
    def doUpdateDeviceTreeFromMasterServer(self, SynchMasterServerUrl, 
        SynchMasterServerUsername, SynchMasterServerPassword,  REQUEST=None):
        """Update device tree"""

        # Clean up path
        path = SynchMasterServerUrl.strip().strip('/')
        # Add auth info
        path = "%s/wurfl_xml_diff_tree_raw?__ac_name=%s&amp;__ac_password=%s" \
            % (path, SynchMasterServerUsername.strip(), SynchMasterServerPassword.strip())
        # Retrieve XML    
        f = urllib2.urlopen(path)
        xml = f.read()
        
        if not xml:
            if REQUEST is not None:
                REQUEST.RESPONSE.redirect(self.absolute_url() 
                    + '/wurfl_import_xml?portal_status_message=Error%20retrieving%20XML')
            return False
            
        res = self.doUpdateDeviceTreeXML(wurflfile=StringIO(xml), 
            update_existing_devices=True)
            
        if res:
            if REQUEST is not None:
                REQUEST.RESPONSE.redirect(self.absolute_url() 
                    + '/wurfl_import_xml?portal_status_message=Success')
            return True
            
        if REQUEST is not None:
            REQUEST.RESPONSE.redirect(self.absolute_url() 
                + '/wurfl_import_xml?portal_status_message=Operation%20failed')
        return False
        
    security.declareProtected(CMFCorePermissions.ManagePortal, 'doSaveMasterServerSettings')
    def doSaveMasterServerSettings(self, SynchMasterServerUrl, 
        SynchMasterServerUsername, SynchMasterServerPassword,  REQUEST=None):
        """Save synch settings"""

        self.edit(SynchMasterServerUrl=SynchMasterServerUrl,
            SynchMasterServerUsername=SynchMasterServerUsername,
            SynchMasterServerPassword=SynchMasterServerPassword)
            
        if REQUEST is not None:
             REQUEST.RESPONSE.redirect(self.absolute_url() 
                    + '/wurfl_import_xml?portal_status_message=Settings%20saved')

    def getGroups(self):
        se = getToolByName(self, 'schema_editor_tool')
        return se.atse_getSchemataNames('MobileDevice')
        
    def getGroupCapabilities(self, group):
        schematas = self.Schemata()
        fieldset = schematas[group]
        return [field.getName() for field in fieldset.fields()]
        
    def getGroupCapabilityFields(self, group):
        schematas = self.Schemata()
        fieldset = schematas[group]
        return fieldset.fields()
   
    def getChildren(self, sort=False, batch_start=0, batch_size=0):
        pc = getToolByName(self, 'portal_catalog')

        brains = pc(portal_type='MobileDevice',
                    path={'query':"/".join(self.getPhysicalPath()), 'depth':1},
                    sort_on=sort and 'sortable_title' or None)
       
        if batch_size:
            brains_batch = brains[batch_start:batch_start+batch_size]

        return [b.getObject() for b in brains_batch]

    security.declareProtected(CMFCorePermissions.ModifyPortalContent, 'setAttribute')
    def setAttribute(self, attribute, value):
        old_value = self.get(attribute)
        self.getField(attribute).set(self, value)
        if old_value != self.get(attribute):
            # Flag as modified
            modified_fields = list(self.getModifiedFields())
            if not attribute in modified_fields:
                self.setModifiedFields(modified_fields + [attribute])
            self.reindexObject()

    security.declareProtected(CMFCorePermissions.ModifyPortalContent, 'unFlagModified')
    def unFlagModified(self, attribute):
        modified_fields = list(self.getModifiedFields())
        li = []
        for m in modified_fields:
            if m != attribute:
                li.append(m)
        self.setModifiedFields(li)
        self.reindexObject()

    def getActualDeviceRoot(self):
        """
        If self is not an actual device root then traverse up until
        one is found
        """
        if self.getIsActualDeviceRoot():
            return self
        parent = self.aq_parent           
        while getattr(parent, 'portal_type', '') == 'MobileDevice':
            if parent.getIsActualDeviceRoot():
                return parent
            parent = parent.aq_parent
            
        return None

def modify_fti(fti):
    for a in fti['actions']:
        if a['id'] in ('view',):
            a['visible'] = 0
    return fti

registerType(MobileDevice, PROJECTNAME)
